-- by mor2000

--------------------
-- UPK_FillTrigger (fills trailers and/or shovels with specific fillType)

local UPK_FillTrigger_mt = ClassUPK(UPK_FillTrigger,UniversalProcessKit)
InitObjectClass(UPK_FillTrigger, "UPK_FillTrigger")
UniversalProcessKit.addModule("filltrigger",UPK_FillTrigger)

UPK_FillTrigger.EVENT_FILLTYPESELECTED = UniversalProcessKitEvent.getNextEventId()
UPK_FillTrigger.EVENT_FILLINGSTOPPED = UniversalProcessKitEvent.getNextEventId()
UPK_FillTrigger.EVENT_ISACTIVE = UniversalProcessKitEvent.getNextEventId()

function UPK_FillTrigger:new(nodeId, parent)
	printFn('UPK_FillTrigger:new(',nodeId,', ',parent,')')
	local self = UniversalProcessKit:new(nodeId, parent, UPK_FillTrigger_mt)
	registerObjectClassName(self, "UPK_FillTrigger")
	
	self.fillFillType=nil
	local fillFillTypeStr = getStringFromUserAttribute(nodeId, "fillType")
	if fillFillTypeStr~=nil then
		self.fillFillType = UniversalProcessKit.fillTypeNameToInt[fillFillTypeStr]
		if self.fillFillType==nil then
			printInfo('unknown fillType "',fillFillTypeStr,'"')
		end
	end
	
	self.fillFillTypes={}
	self.useFillFillTypes=false
	local fillTypesArr = getArrayFromUserAttribute(nodeId, "fillTypes", false)
	if fillTypesArr~=false then
		local fillTypesArrLen = length(fillTypesArr)
		for i=1,fillTypesArrLen do
			local fillType = UniversalProcessKit.fillTypeNameToInt[fillTypesArr[i]]
			if fillType~=nil then
				table.insert(self.fillFillTypes,fillType)
				self.useFillFillTypes=true
			else
				printInfo('unknown fillType "',fillTypesArr[i],'"')
			end
		end
	end
	
	if self.useFillFillTypes and length(self.fillFillTypes)==1 then
		printInfo('just 1 filltype found in "fillTypes"')
		self.fillFillType=self.fillFillTypes[1]
		self.useFillFillTypes=false
		self.fillFillTypes={}
	end
	
	self:printInfo('self.useFillFillTypes: ',self.useFillFillTypes)
	
	if self.fillFillType~=nil and self.useFillFillTypes then
		printInfo('use either "fillType" or "fillTypes" - merging both for now')
		table.insert(self.fillFillTypes,self.fillFillType,1)
		self.fillFillType=nil
	end
	
	self.vehiclesToFill = {}
	self.isFilling = false
	self.isActive = false

	-- activateInput, activateInputBinding
	local activateInput = getStringFromUserAttribute(nodeId, "activateInput", nil)
	if activateInput==nil and self.useFillFillTypes then
		self:printInfo('always use activateInput for multiple fillTypes - using "ACTIVATE_OBJECT" for now')
		activateInput = 'ACTIVATE_OBJECT'
	end
		
	self.useActivateInput = activateInput~=nil
	
	if activateInput~=nil then
		if not InputBinding[activateInput] then
			self:printErr('unknown input "',activateInput,'" - using "ACTIVATE_OBJECT" for now')
			self.activateInputBinding = 'ACTIVATE_OBJECT'
		else
			self.activateInputBinding = activateInput
		end
	end

	-- start and stopFillingText
	local startFillingText = self.i18n[getStringFromUserAttribute(nodeId, "startFillingText")]
	local stopFillingText = self.i18n[getStringFromUserAttribute(nodeId, "stopFillingText")]
	self:printInfo('startFillingText ',startFillingText)
	self:printInfo('stopFillingText ',stopFillingText)
	self.startFillingText = returnNilIfEmptyString(startFillingText) or self.i18n['action_siloStartFilling']
	self.stopFillingText = returnNilIfEmptyString(stopFillingText) or self.i18n['action_siloStopFilling']

	-- stationName
	local stationName = self.name
	if self.i18nNameSpace~=nil and self.i18nNameSpace~="" and type(ModsUtil['modNameToMod'][self.i18nNameSpace])=="table" then
		stationName = ModsUtil['modNameToMod'][self.i18nNameSpace].title
	end
	self.stationName = returnNilIfEmptyString(self.i18n[getStringFromUserAttribute(nodeId, "stationName")]) or stationName
	
    self.fillLitersPerSecond = getNumberFromUserAttribute(nodeId, "fillLitersPerSecond", 1500, 0)
	self.createFillType = getBoolFromUserAttribute(nodeId, "createFillType", false)
    self.pricePerLiter = getNumberFromUserAttribute(nodeId, "pricePerLiter", 0)
	
	self.preferMapDefaultPrice = getBoolFromUserAttribute(nodeId, "preferMapDefaultPrice", false)
	self.pricePerLiterMultiplier = getVectorFromUserAttribute(nodeId, "pricePerLiterMultiplier", "1 1 1")
	self.pricesPerLiter = {}
	
	self.fillOnlyWholeNumbers = getBoolFromUserAttribute(nodeId, "fillOnlyWholeNumbers", false)
	self.vehicleToAmountNotYetFilled = {}
	
	-- money change id
	
	self.moneyChangeId = getMoneyTypeId()
	self.vehicleToMoneyChanged = {}
	
	-- statName
	
	self.statName = getStatNameFromUserAttribute(nodeId)

	-- auto allow...

	local fillTypes = {}
	if self.useFillFillTypes==false then
		table.insert(fillTypes,self.fillFillType or self:getFillType())
	else
		fillTypes=self.fillFillTypes
	end
	
	self.allowedVehicles={}
	self.allowedVehicles[UniversalProcessKit.VEHICLE_TIPPER] = getBoolFromUserAttribute(nodeId, "allowTipper", true)
	self.allowedVehicles[UniversalProcessKit.VEHICLE_SHOVEL] = getBoolFromUserAttribute(nodeId, "allowShovel", true)
	
	self.allowedVehicles[UniversalProcessKit.VEHICLE_SOWINGMACHINE] = getBoolFromUserAttribute(nodeId, "allowSowingMachine", isInTable(fillTypes,Fillable.FILLTYPE_SEEDS))
	self.allowedVehicles[UniversalProcessKit.VEHICLE_WATERTRAILER] = getBoolFromUserAttribute(nodeId, "allowWaterTrailer", isInTable(fillTypes,UniversalProcessKit.FILLTYPE_WATER))
	self.allowedVehicles[UniversalProcessKit.VEHICLE_MILKTRAILER] = getBoolFromUserAttribute(nodeId, "allowMilkTrailer", isInTable(fillTypes,UniversalProcessKit.FILLTYPE_MILK))
	self.allowedVehicles[UniversalProcessKit.VEHICLE_LIQUIDMANURETRAILER] = getBoolFromUserAttribute(nodeId, "allowLiquidManureTrailer", isInTable(fillTypes,UniversalProcessKit.FILLTYPE_LIQUIDMANURE))
	self.allowedVehicles[UniversalProcessKit.VEHICLE_SPRAYER] = getBoolFromUserAttribute(nodeId, "allowSprayer", isInTable(fillTypes,UniversalProcessKit.FILLTYPE_FERTILIZER))
	self.allowedVehicles[UniversalProcessKit.VEHICLE_MANURESPREADER] = getBoolFromUserAttribute(nodeId, "allowManureSpreader", isInTable(fillTypes,UniversalProcessKit.FILLTYPE_MANURE))
	
	self.allowedVehicles[UniversalProcessKit.VEHICLE_FUELTRAILER] = getBoolFromUserAttribute(nodeId, "allowFuelTrailer", isInTable(fillTypes,UniversalProcessKit.FILLTYPE_FUEL))
	self.allowedVehicles[UniversalProcessKit.VEHICLE_MOTORIZED] = getBoolFromUserAttribute(nodeId, "allowMotorized", false)
	
	self.allowPallets = getBoolFromUserAttribute(nodeId, "allowPallets", self.palletFilename~=nil)

	self.allowWalker = getBoolFromUserAttribute(nodeId, "allowWalker", false)
	
	self:addTrigger()
	
	-- actions
	self:getActionUserAttributes('IfFilling')
	UniversalProcessKit.syncActionThruStream('IfFilling')
	
	self:getActionUserAttributes('IfFillingStarted')
	self:getActionUserAttributes('IfFillingStopped')
	
	self:printFn('UPK_FillTrigger:new done')
	
	return self
end

function UPK_FillTrigger:delete()
	self:printFn('UPK_FillTrigger:delete()')
	UniversalProcessKitListener.removeUpdateable(self)
	if self.useActivateInput then
		UniversalProcessKitListener.unregisterKeyFunction(self.activateInputBinding,self)
	end
	UPK_FillTrigger:superClass().delete(self)
end

function UPK_FillTrigger:postLoad()
	self:printFn('UPK_FillTrigger:postLoad()')
	-- default: stopped
	self:operateAction('IfFillingStopped',nil,true)
	
	UPK_FillTrigger:superClass().postLoad(self)
end

function UPK_FillTrigger:eventCallback(eventType,...)
	self:printFn('UPK_FillTrigger:eventCallBack(',eventType,'...)')
	if eventType==UPK_FillTrigger.EVENT_FILLTYPESELECTED then
		-- like UPK_FillTrigger:onFillTypeSelection
		local selectedFillType = ...
		self:onFillTypeSelection(selectedFillType,true)
	elseif eventType==UPK_FillTrigger.EVENT_FILLINGSTOPPED then
		self:onFillingStopped(true)
	end
end

function UPK_FillTrigger:writeStream(streamId, connection)
	self:printFn('UPK_FillTrigger:writeStream(',streamId,', ',connection,')')
	UPK_FillTrigger:superClass().writeStream(self, streamId, connection)
	if not connection:getIsServer() then
		self:printInfo('write to stream self.isActive = ',self.isActive)
		streamWriteAuto(streamId,self.isActive)
		self:printInfo('write to stream self.fillFillType = ',self.fillFillType)
		streamWriteAuto(streamId,self.fillFillType)
	end
end;

function UPK_FillTrigger:readStream(streamId, connection)
	self:printFn('UPK_FillTrigger:readStream(',streamId,', ',connection,')')
	UPK_FillTrigger:superClass().readStream(self, streamId, connection)
	if connection:getIsServer() then
		self.isActive = streamReadAuto(streamId)
		self:printInfo('read from stream self.isActive = ',self.isActive)
		self.fillFillType = streamReadAuto(streamId)
		self:printInfo('read from stream self.fillFillType = ',self.fillFillType)
	end
end;

function UPK_FillTrigger:loadExtraNodes(xmlFile, key)
	self:printFn('UPK_FillTrigger:loadExtraNodes(',xmlFile,', ',key,')')
	if self.useFillFillTypes then
		local selectedFillType = getXMLString(xmlFile, key .. "#selectedFillType")
		if selectedFillType~=nil then
			local fillType=UniversalProcessKit.fillTypeNameToInt[selectedFillType]
			if fillType~=nil then
				self.fillFillType=fillType
			end
		end
	end
	return true
end

function UPK_FillTrigger:getSaveExtraNodes(nodeIdent)
	self:printFn('UPK_FillTrigger:getSaveExtraNodes(',nodeIdent,')')
	local nodes=""
	if self.useFillFillTypes and self.fillFillType~=nil then
		nodes=nodes .. ' selectedFillType="'..UniversalProcessKit.fillTypeIntToName[self.fillFillType]..'"'
	end
	return nodes
end

function UPK_FillTrigger:triggerUpdate(tool,isInTrigger)
	self:printFn('UPK_FillTrigger:triggerUpdate(',tool,', ',isInTrigger,')')
	if self.isEnabled then
		local hasFillableEntitiesBefore = length(self.vehiclesToFill)>0
		
		-- check for vehicles
		for vehicleType,isAllowed in pairs(self.allowedVehicles) do
			if isAllowed and UniversalProcessKit.isVehicleType(tool, vehicleType) then
				if isInTrigger then
					if not isInTable(self.vehiclesToFill,tool) then
						table.insert(self.vehiclesToFill,1,tool)
					end
					self.vehicleToAmountNotYetFilled[tool] = 0
				else
					removeValueFromTable(self.vehiclesToFill,tool)
					self.vehicleToAmountNotYetFilled[tool] = nil
					if self.vehicleToMoneyChanged[tool]~=nil and self.vehicleToMoneyChanged[tool]~=0 then
						g_currentMission:addMoneyChange(-self.vehicleToMoneyChanged[tool], self.moneyChangeId);
						g_currentMission:showMoneyChange(self.moneyChangeId, self.i18n["finance_"..self.statName])
						self.vehicleToMoneyChanged[tool] = nil
					end
				end
			end
		end
		
		-- check also for pallets
		if self.allowPallets and getIsPallet(tool) then
			if isInTrigger then
				if not isInTable(self.vehiclesToFill,tool) then
					table.insert(self.vehiclesToFill,tool)
				end
				self.vehicleToAmountNotYetFilled[tool] = 0
			else
				removeValueFromTable(self.vehiclesToFill,tool)
				self.vehicleToAmountNotYetFilled[tool] = nil
				if self.vehicleToMoneyChanged[tool]~=nil and self.vehicleToMoneyChanged[tool]~=0 then
					g_currentMission:addMoneyChange(-self.vehicleToMoneyChanged[tool], self.moneyChangeId);
					g_currentMission:showMoneyChange(self.moneyChangeId, self.i18n["finance_"..self.statName])
					self.vehicleToMoneyChanged[tool] = nil
				end
			end
		end
		
		local hasFillableEntitiesAfter = length(self.vehiclesToFill)>0
		if not hasFillableEntitiesBefore and hasFillableEntitiesAfter then
			if self.useActivateInput then
				UniversalProcessKitListener.registerKeyFunction(self.activateInputBinding,self,'inputCallback',self.startFillingText)
			end
			UniversalProcessKitListener.addUpdateable(self)
		elseif hasFillableEntitiesBefore and not hasFillableEntitiesAfter then
			if self.useActivateInput then
				UniversalProcessKitListener.unregisterKeyFunction(self.activateInputBinding,self)
			end
			UniversalProcessKitListener.removeUpdateable(self)
			if self.isActive then
				self:onFillingStopped(true)
			end
			self.lastVehicleToFill = nil
		end
	end
end

function UPK_FillTrigger:update(dt)
	self:printAll('UPK_FillTrigger:update(',dt,')')
	
	self:printInfo('isEnabled ',self.isEnabled,', isActive ',self.isActive)
	
	if self.isEnabled and (not self.useActivateInput or self.isActive) then
		-- vehicle to fill
		local vehicle = nil
		
		if self.lastVehicleToFill~=nil then
			vehicle = self.lastVehicleToFill
		else
			-- find new vehicle
			for _,v in ipairs(self.vehiclesToFill) do
				if v:getIsActiveForInput(false) then
					vehicle = v
					break
				end
			end
		end
		
		if vehicle==nil or not isInTable(self.vehiclesToFill,vehicle) then
			self:onFillingStopped()
			return
		end

		local added = 0
		
		self.lastVehicleToFill = vehicle

		local deltaFillLevel = floor(self.fillLitersPerSecond * 0.001 * dt,8)
		self:printInfo('trying to fill ',deltaFillLevel)
		for k,v in pairs(self.allowedVehicles) do
			if v and UniversalProcessKit.isVehicleType(vehicle, k) then
				self:printAll('vehicle allowed')
				if k==UniversalProcessKit.VEHICLE_MIXERWAGONPICKUP then
					added = self:fillMixerWagonPickup(vehicle, deltaFillLevel)
				elseif k==UniversalProcessKit.VEHICLE_MIXERWAGONTRAILER then
					added = self:fillMixerWagonTrailer(vehicle, deltaFillLevel)
				elseif ((k==UniversalProcessKit.VEHICLE_TIPPER and not UniversalProcessKit.isVehicleType(vehicle, UniversalProcessKit.VEHICLE_MIXERWAGONTRAILER)) or
						(k==UniversalProcessKit.VEHICLE_SHOVEL and not UniversalProcessKit.isVehicleType(vehicle, UniversalProcessKit.VEHICLE_MIXERWAGONPICKUP)) or
						k==UniversalProcessKit.VEHICLE_SOWINGMACHINE or
						k==UniversalProcessKit.VEHICLE_WATERTRAILER or
						k==UniversalProcessKit.VEHICLE_MANURESPREADER or
						k==UniversalProcessKit.VEHICLE_MILKTRAILER or
						k==UniversalProcessKit.VEHICLE_LIQUIDMANURETRAILER or
						k==UniversalProcessKit.VEHICLE_SPRAYER or
						k==UniversalProcessKit.VEHICLE_FUELTRAILER) then
					added = self:fillTrailer(vehicle, deltaFillLevel)
				elseif k==UniversalProcessKit.VEHICLE_MOTORIZED then
					added = self:fillMotorized(vehicle, deltaFillLevel)
				end
				break
			end
		end
		if self.allowPallets and getIsPallet(vehicle) and vehicle.setFillLevel~=nil then
			added = self:fillPallet(vehicle, deltaFillLevel)
		end
		
		-- is filling
		if round(added,8)>0 then
			self:setIsFilling(true)
			self:operateAction('IfFilling',added)
		else
			self:onFillingStopped()
		end
	end
end

function UPK_FillTrigger:getFillLevel(fillType)
	self:printFn('UPK_FillTrigger:getFillLevel(',fillType,')')
	return UPK_FillTrigger:superClass().getFillLevel(self, fillType or self.fillFillType or self:getFillType()) or 0
end

function UPK_FillTrigger:getPricePerLiter(fillType)
	self:printFn('UPK_FillTrigger:getPricePerLiter(',fillType,')')
	fillType = fillType or self.fillFillType
	if self.pricesPerLiter[fillType]~=nil then
		return self.pricesPerLiter[fillType]
	end
	local pricePerLiter = self.pricePerLiter
	if self.preferMapDefaultPrice then
		pricePerLiter = g_currentMission.economyManager:getPricePerLiter(fillType) or pricePerLiter
	end
	pricePerLiter = pricePerLiter * (self.pricePerLiterMultiplier[g_currentMission.missionStats.difficulty] or 1)
	self.pricesPerLiter[fillType] = pricePerLiter
	return pricePerLiter
end

function UPK_FillTrigger:inputCallback(inputName)
	self:printFn('UPK_FillTrigger:inputCallback(',inputName,')')
	if self.activateInputBinding==inputName then
		if self.useFillFillTypes and not self.isActive then
			local fillLevels = {}
			for _,fillType in pairs(self.fillFillTypes) do
				if self.createFillType then
					fillLevels[fillType] = 99999
				else
					fillLevels[fillType] = self:getFillLevel(fillType)
				end
			end
			local capacity = self:getCapacity()
			g_gui:showSiloDialog( {
				title = self.stationName,
				fillLevels = fillLevels,
				capacity = capacity,
				callback = self.onFillTypeSelection,
				target = self})
		else
			self:onFillingStopped()
		end
	end
end

function UPK_FillTrigger:onFillTypeSelection(selectedFillType,alreadySent)
	self:printFn('UPK_FillTrigger:onFillTypeSelection(',selectedFillType,', ',alreadySent,')')
	if not self.isActive and selectedFillType~=UniversalProcessKit.FILLTYPE_UNKNOWN then
		self.fillFillType = selectedFillType
		self:setIsActive(true)
		if not alreadySent then
			self:sendEvent(UPK_FillTrigger.EVENT_FILLTYPESELECTED,self.fillFillType)
		end
	end
end

function UPK_FillTrigger:setIsActive(isActive)
	self:printFn('UPK_FillTrigger:setIsActive(',isActive,')')
	if self.useActivateInput then
		if isActive and not self.isActive then
			UniversalProcessKitListener.updateKeyFunctionDisplayText(self.activateInputBinding,self,self.stopFillingText)
			self.isActive = true
		elseif not isActive and self.isActive then
			UniversalProcessKitListener.updateKeyFunctionDisplayText(self.activateInputBinding,self,self.startFillingText)
			self.isActive = false
			self.lastVehicleToFill = nil
		end
	end
end

function UPK_FillTrigger:setIsFilling(isFilling,alreadySent)
	self:printFn('UPK_FillTrigger:setIsFilling(',isFilling,', ',alreadySent,')')
	if isFilling and not self.isFilling then
		self:operateAction('IfFillingStarted')
		self.isFilling = true
	elseif not isFilling and self.isFilling then
		self:operateAction('IfFillingStopped')
		self.isFilling = false
		if not alreadySent then
			self:sendEvent(UPK_FillTrigger.EVENT_FILLINGSTOPPED)
		end
	end
end

function UPK_FillTrigger:onFillingStopped(alreadySent)
	self:printFn('UPK_FillTrigger:onFillingStopped(',alreadySent,')')
	self:setIsActive(false)
	self:setIsFilling(false)

	for _,vehicle in ipairs(self.vehiclesToFill) do
		if self.vehicleToMoneyChanged[vehicle]~=nil and self.vehicleToMoneyChanged[vehicle]~=0 then
			g_currentMission:addMoneyChange(-self.vehicleToMoneyChanged[vehicle], self.moneyChangeId);
			g_currentMission:showMoneyChange(self.moneyChangeId, self.i18n["finance_"..self.statName])
			self.vehicleToMoneyChanged[vehicle] = nil
		end
	end
end

function UPK_FillTrigger:fillTrailer(trailer, deltaFillLevel) -- tippers, shovels etc
	self:printFn('UPK_FillTrigger:fillTrailer(',trailer,', ',deltaFillLevel,')')
	if self.isServer and self.isEnabled then
		local fillFillType = self.fillFillType or self:getFillType() -- for single, fifo and filo
		self:printAll('fillFillType ',fillFillType)
		if fillFillType~=UniversalProcessKit.FILLTYPE_UNKNOWN then
			if not trailer:allowFillType(fillFillType,false) then
				self:printAll('trailer does not accept ',fillFillType)
				return 0
			end
			
			if trailer:getFreeCapacity(fillFillType)<0.0001 then
				self.printAll('trailer is full of ',fillFillType)
				return 0
			end
			
			local fillLevel = self:getFillLevel(fillFillType)
			if fillLevel<0.0001 and not self.createFillType then
				self:printAll('filltrigger has nothing to fill')
				return 0
			end

			if self.fillOnlyWholeNumbers then
				self.vehicleToAmountNotYetFilled[trailer] = (self.vehicleToAmountNotYetFilled[trailer] or 0) - deltaFillLevel
				deltaFillLevel = -mathfloor(self.vehicleToAmountNotYetFilled[trailer])
				self.vehicleToAmountNotYetFilled[trailer] = self.vehicleToAmountNotYetFilled[trailer] + deltaFillLevel
			end
			
			if not self.createFillType then
				deltaFillLevel = math.min(deltaFillLevel, fillLevel)
			end
			
			if deltaFillLevel==0 then
				return 0
			end
			self:printAll('deltaFillLevel is ',deltaFillLevel)
			
			local trailerFillLevel = trailer:getFillLevel(fillFillType)
			self:printAll('trailerFillLevel is ',trailerFillLevel)
			trailer:setFillLevel(round(trailerFillLevel + deltaFillLevel,8), fillFillType)
			deltaFillLevel = trailer:getFillLevel(fillFillType) - trailerFillLevel
			self:printAll('deltaFillLevel is ',deltaFillLevel)
				
			if deltaFillLevel~=0 then
				if not self.createFillType then
					deltaFillLevel=-self:addFillLevel(-deltaFillLevel,fillFillType)
					self:printAll('removed from filltrigger ',deltaFillLevel)
				end
				local pricePerLiter = self:getPricePerLiter(fillFillType)
				if pricePerLiter~=0 then
					local price = deltaFillLevel * pricePerLiter
					self.vehicleToMoneyChanged[trailer] = (self.vehicleToMoneyChanged[trailer] or 0) + price
				end
				return deltaFillLevel
			end
		end
	end
	return 0
end

function UPK_FillTrigger:fillMotorized(trailer, deltaFillLevel) -- motorized
	self:printFn('UPK_FillTrigger:fillMotorized(',trailer,', ',deltaFillLevel,')')
	if self.isServer and self.isEnabled then
		local fillFillType = self.fillFillType or self:getFillType() -- for single, fifo and filo
		self:printAll('fillFillType ',fillFillType)
		if fillFillType==UniversalProcessKit.FILLTYPE_FUEL then
			
			if self.fillOnlyWholeNumbers then
				self.vehicleToAmountNotYetFilled[trailer] = (self.vehicleToAmountNotYetFilled[trailer] or 0) - deltaFillLevel
				deltaFillLevel = -mathfloor(self.vehicleToAmountNotYetFilled[trailer])
				self.vehicleToAmountNotYetFilled[trailer] = self.vehicleToAmountNotYetFilled[trailer] + deltaFillLevel
			end
			
			if deltaFillLevel==0 then
				return 0
			end
			
			local trailerFillLevel = trailer.fuelFillLevel
			self:printAll('trailerFillLevel ',trailerFillLevel)
			local fillLevel = self:getFillLevel(fillFillType)
			self:printAll('fillLevel ',fillLevel)
			if (fillLevel>0 or self.createFillType) and round(trailerFillLevel,1)<round(trailer.fuelCapacity,1) then
				if not self.createFillType then
					deltaFillLevel=math.min(deltaFillLevel, fillLevel)
				end
				trailer:setFuelFillLevel(round(trailerFillLevel + deltaFillLevel,8))
				deltaFillLevel = trailer.fuelFillLevel - trailerFillLevel
				if deltaFillLevel~=0 then
					if not self.createFillType then
						deltaFillLevel=-self:addFillLevel(-deltaFillLevel, fillFillType)
					end
					local pricePerLiter = self:getPricePerLiter(fillFillType)
					if pricePerLiter~=0 then
						local price = deltaFillLevel * pricePerLiter
						self.vehicleToMoneyChanged[trailer] = self.vehicleToMoneyChanged[trailer] - price
					end
				end
				return deltaFillLevel
			end
		end
	end
	return 0
end

function UPK_FillTrigger:fillMixerWagonPickup(trailer, deltaFillLevel) -- mixing wagon pickups etc
	self:printFn('UPK_FillTrigger:fillMixerWagonPickup(',trailer,', ',deltaFillLevel,')')
	self:printAll('trailer.isTurnedOn ',trailer.isTurnedOn)
	if self.isServer and self.isEnabled and trailer.isTurnedOn==true then
		local fillFillType = self.fillFillType or self:getFillType() -- for single, fifo and filo
		self:printAll('fillFillType ',fillFillType)
		if fillFillType~=UniversalProcessKit.FILLTYPE_UNKNOWN then
			
			if self.fillOnlyWholeNumbers then
				self.vehicleToAmountNotYetFilled[trailer] = (self.vehicleToAmountNotYetFilled[trailer] or 0) - deltaFillLevel
				deltaFillLevel = -mathfloor(self.vehicleToAmountNotYetFilled[trailer])
				self.vehicleToAmountNotYetFilled[trailer] = self.vehicleToAmountNotYetFilled[trailer] + deltaFillLevel
			end
			
			if deltaFillLevel==0 then
				return 0
			end
			
			local trailerFillLevel = trailer:getFillLevel(fillFillType)
			local fillLevel = self:getFillLevel(fillFillType)
			--self:print('fillLevel '..tostring(fillLevel))
			--self:print('trailer:allowFillType(fillFillType, false) '..tostring(trailer:allowFillType(fillFillType, false)))
			
			if (fillLevel>0 or self.createFillType) and
				trailer:allowFillType(fillFillType, false) and
				trailer.fillLevel<trailer.capacity then
				
				trailer:resetFillLevelIfNeeded(fillFillType)
				trailerFillLevel = trailer:getFillLevel(fillFillType)
				if not self.createFillType then
					deltaFillLevel=math.min(deltaFillLevel, fillLevel)
				end
				trailer:setFillLevel(round(trailerFillLevel + deltaFillLevel,8), fillFillType)
				deltaFillLevel = trailer:getFillLevel(fillFillType) - trailerFillLevel
				if deltaFillLevel~=0 then
					if not self.createFillType then
						deltaFillLevel=-self:addFillLevel(-deltaFillLevel,fillFillType)
					end
					local pricePerLiter = self:getPricePerLiter(fillFillType)
					if pricePerLiter~=0 then
						local price = deltaFillLevel * pricePerLiter
						self.vehicleToMoneyChanged[trailer] = self.vehicleToMoneyChanged[trailer] - price
					end
					trailer.mixerWagonLastPickupTime = trailer.time
					return deltaFillLevel
				end
			end
		end
	end
	return 0
end

function UPK_FillTrigger:fillMixerWagonTrailer(trailer, deltaFillLevel) -- mixer wagon itself etc
	self:printFn('UPK_FillTrigger:fillMixerWagonTrailer(',trailer,', ',deltaFillLevel,')')
	if self.isServer and self.isEnabled then
		local fillFillType = self.fillFillType or self:getFillType() -- for single, fifo and filo
		self:printAll('fillFillType ',fillFillType)
		if fillFillType~=UniversalProcessKit.FILLTYPE_UNKNOWN then
			
			if self.fillOnlyWholeNumbers then
				self.vehicleToAmountNotYetFilled[trailer] = (self.vehicleToAmountNotYetFilled[trailer] or 0) - deltaFillLevel
				deltaFillLevel = -mathfloor(self.vehicleToAmountNotYetFilled[trailer])
				self.vehicleToAmountNotYetFilled[trailer] = self.vehicleToAmountNotYetFilled[trailer] + deltaFillLevel
			end
			
			if deltaFillLevel==0 then
				return 0
			end
			
			local trailerFillLevel = trailer:getFillLevel(fillFillType)
			local fillLevel = self:getFillLevel(fillFillType)
			--self:print('fillLevel '..tostring(fillLevel))
			--self:print('trailer:allowFillType(fillFillType, false) '..tostring(trailer:allowFillType(fillFillType, false)))
			
			if (fillLevel>0 or self.createFillType) and
				trailer:allowFillType(fillFillType, false) and
				trailer.fillLevel<trailer.capacity then
				
				trailer:resetFillLevelIfNeeded(fillFillType)
				trailerFillLevel = trailer:getFillLevel(fillFillType)
				if not self.createFillType then
					deltaFillLevel=math.min(deltaFillLevel, fillLevel)
				end
				trailer:setFillLevel(round(trailerFillLevel + deltaFillLevel,8), fillFillType)
				deltaFillLevel = trailer:getFillLevel(fillFillType) - trailerFillLevel
				if deltaFillLevel~=0 then
					if not self.createFillType then
						deltaFillLevel=-self:addFillLevel(-deltaFillLevel,fillFillType)
					end
					local pricePerLiter = self:getPricePerLiter(fillFillType)
					if pricePerLiter~=0 then
						local price = deltaFillLevel * pricePerLiter
						self.vehicleToMoneyChanged[trailer] = self.vehicleToMoneyChanged[trailer] - price
					end
					trailer.mixerWagonLastPickupTime = trailer.time
					return deltaFillLevel
				end
			end
		end
	end
	return 0
end

function UPK_FillTrigger:fillPallet(trailer, deltaFillLevel) -- pallets
	self:printFn('UPK_FillTrigger:fillPallet(',trailer,', ',deltaFillLevel,')')
	if self.isServer and self.isEnabled then
		local fillFillType = self.fillFillType or self:getFillType() -- for single, fifo and filo
		self:printAll('fillFillType '..tostring(fillFillType))
		if fillFillType~=UniversalProcessKit.FILLTYPE_UNKNOWN then
			
			if self.fillOnlyWholeNumbers then
				self.vehicleToAmountNotYetFilled[trailer] = (self.vehicleToAmountNotYetFilled[trailer] or 0) - deltaFillLevel
				deltaFillLevel = -mathfloor(self.vehicleToAmountNotYetFilled[trailer])
				self.vehicleToAmountNotYetFilled[trailer] = self.vehicleToAmountNotYetFilled[trailer] + deltaFillLevel
			end
			
			if deltaFillLevel==0 then
				return 0
			end
			
			local trailerFillType = trailer:getFillType()
			local trailerFillLevel = trailer:getFillLevel()
			local fillLevel = self:getFillLevel(fillFillType)
			--self:print('fillLevel '..tostring(fillLevel))
			--self:print('trailer:allowFillType(fillFillType, false) '..tostring(trailer:allowFillType(fillFillType, false)))
			if (fillLevel>0 or self.createFillType) and
				fillFillType==trailerFillType and
				trailerFillLevel<trailer.capacity then
				
				if not self.createFillType then
					deltaFillLevel=math.min(deltaFillLevel, fillLevel)
				end
				trailer:setFillLevel(round(trailerFillLevel + deltaFillLevel,8), fillFillType)
				deltaFillLevel = trailer:getFillLevel() - trailerFillLevel
				if deltaFillLevel~=0 then
					if not self.createFillType then
						deltaFillLevel=-self:addFillLevel(-deltaFillLevel,fillFillType)
					end
					local pricePerLiter = self:getPricePerLiter(fillFillType)
					if pricePerLiter~=0 then
						local price = deltaFillLevel * pricePerLiter
						self.vehicleToMoneyChanged[trailer] = self.vehicleToMoneyChanged[trailer] - price
					end
					return deltaFillLevel
				end
			end
		end
	end
	return 0
end

function UPK_FillTrigger:getIsActivatable(trailer)
	self:printFn('UPK_FillTrigger:getIsActivatable(',trailer,')')
	local fillFillType = self.fillFillType or self:getFillType()
	if trailer:allowFillType(fillFillType, false) and
		(self:getFillLevel(fillFillType)>0 or self.createFillType) then
		return true
	end
	return false
end

function UPK_FillTrigger:referMoneyToVehicle(vehicle,moneyChanged)
	self.vehicleToMoneyChanged[vehicle] = (self.vehicleToMoneyChanged[vehicle] or 0) + moneyChanged
end
