-- by mor2000

--------------------
-- UPK_ObjectTrigger (reacts to objects like pallets, bales, bigbags, basketballs, .. you name it)

local UPK_ObjectTrigger_mt = ClassUPK(UPK_ObjectTrigger, UniversalProcessKit)
InitObjectClass(UPK_ObjectTrigger, "UPK_ObjectTrigger")
UniversalProcessKit.addModule("objecttrigger", UPK_ObjectTrigger)

function UPK_ObjectTrigger:new(nodeId, parent)
	printFn('UPK_ObjectTrigger:new(',nodeId,', ',parent,')')
	local self = UniversalProcessKit:new(nodeId, parent, UPK_ObjectTrigger_mt)
	registerObjectClassName(self, "UPK_ObjectTrigger")
	
	local objectTypes = {}
	local objectTypeToClass = {}
	
	-- in-game pallets
	objectTypes["fertilizerTank"] = "$data/objects/pallets/fertilizerTank.i3d"
	objectTypeToClass["fertilizerTank"] = FillablePallet
	objectTypes["palletPoplar"] = "$data/objects/pallets/palletPoplar.i3d"
	objectTypeToClass["palletPoplar"] = FillablePallet
	objectTypes["palletPotato"] = "$data/objects/pallets/palletPotato.i3d"
	objectTypeToClass["palletPotato"] = FillablePallet
	objectTypes["palletSugarBeet"] = "$data/objects/pallets/palletSugarBett.i3d"
	objectTypeToClass["palletSugarBeet"] = FillablePallet
	objectTypes["treeSaplingPallet"] = "$data/objects/pallets/treeSaplingPallet.i3d"
	objectTypeToClass["treeSaplingPallet"] = FillablePallet
	objectTypes["woolPallet"] = "$data/objects/pallets/woolPallet.i3d"
	objectTypeToClass["woolPallet"] = FillablePallet
	
	-- in-game bales
	objectTypes["roundbaleGrass"] = "$data/objects/roundbales/roundbaleGrass_w112_d130.i3d"
	objectTypeToClass["roundbaleGrass"] = Bale
	objectTypes["roundbaleHay"] = "$data/objects/roundbales/roundbaleHay_w112_d130.i3d"
	objectTypeToClass["roundbaleHay"] = Bale
	objectTypes["roundbaleSilage"] = "$data/objects/roundbales/roundbaleSilage_w112_d130.i3d"
	objectTypeToClass["roundbaleSilage"] = Bale
	objectTypes["roundbaleStraw"] = "$data/objects/roundbales/roundbaleStraw_w112_d130.i3d"
	objectTypeToClass["roundbaleStraw"] = Bale
	objectTypes["baleGrass"] = "$data/objects/squarebales/baleGrass240.i3d"
	objectTypeToClass["baleGrass"] = Bale
	objectTypes["baleHay"] = "$data/objects/squarebales/baleHay240.i3d"
	objectTypeToClass["baleHay"] = Bale
	objectTypes["baleStraw"] = "$data/objects/squarebales/baleStraw240.i3d"
	objectTypeToClass["baleStraw"] = Bale
	
	-- in-game bigbags
	objectTypes["bigBagFertilizer"] = "$data/objects/bigBagContainer/bigBagContainerFertilizer.i3d"
	objectTypeToClass["bigBagFertilizer"] = FillablePallet
	objectTypes["bigBagPigFood"] = "$data/objects/bigBagContainer/bigBagContainerPigFood.i3d"
	objectTypeToClass["bigBagPigFood"] = FillablePallet
	objectTypes["bigBagSeeds"] = "$data/objects/bigBagContainer/bigBagContainerSeeds.i3d"
	objectTypeToClass["bigBagSeeds"] = FillablePallet
	
	-- in-game objects
	objectTypes["basketball"] = "$data/objects/basketball/basketball.i3d"
	objectTypeToClass["basketball"] = Basketball
	
	-- object
	
	-- just use attribute "object" with pre-set object types like "roundbaleGrass"
	-- or provide a filename in "object" and a class name in "objectClass" (default: Object)
	
	local objectStr = getStringFromUserAttribute(nodeId, "object", "")
	
	self.objectClass = nil
	
	if objectStr~="" and objectTypes[objectStr]~=nil then
		-- preset object
		self:printInfo('preset object ',objectStr,' chosen')
		self.objectFilename = getLongFilename(objectTypes[objectStr],nil)
		self.objectClass = objectTypeToClass[objectStr]
	else
		-- only objectClass important
		local objectClassname = getStringFromUserAttribute(nodeId, "objectClass", "PhysicsObject")
			
		if _g[objectClassname]~=nil and type(_g[objectClassname])=="table" then
			self.objectClass = _g[objectClassname]
		elseif self.i18nNameSpace~=nil and _g[self.i18nNameSpace][objectClassname]~=nil and type(_g[self.i18nNameSpace][objectClassname])=="table" then
			self.objectClass = _g[self.i18nNameSpace][objectClassname]
		else
			self:printErr('no object class "'..objectClassname..'" found')
			return false
		end
	end
	
	if not (self.objectClass==PhysicsObject or derivesFromClass(self.objectClass,PhysicsObject)) then
		self:printErr('object to raect to is not an PhysicsObject')
		return false
	end
	
	self.isLimitedObjectTypeBale = (self.objectClass==Bale or derivesFromClass(self.objectClass,Bale))
	self.isLimitedObjectTypePallet = (self.objectClass==FillablePallet or derivesFromClass(self.objectClass,FillablePallet))

	-- delete objects
	
	self.deleteObjectOnEnter = getBoolFromUserAttribute(nodeId, "deleteObjectOnEnter", false)
	self.deleteObjectOnLeave = getBoolFromUserAttribute(nodeId, "deleteObjectOnLeave", false)
	
	
	-- auto allow...
	
	self.allowedVehicles = {}
	self.allowWalker = false
	self.allowPallets = getBoolFromUserAttribute(nodeId, "allowPallets", self.isLimitedObjectTypePallet)
	self.allowBales = getBoolFromUserAttribute(nodeId, "allowBales", self.isLimitedObjectTypeBale)
	self.allowObjects = getBoolFromUserAttribute(nodeId, "allowObjects", true)
	-- self.allowWood ??
	
	self:addTrigger()
	
	-- end
	
	self:printFn('UPK_ObjectTrigger:new done')
	
	return self
end

function UPK_ObjectTrigger:fitCollisionMaskToAllowedVehicles()
	-- overwrite default function
	self:printFn('UPK_ObjectTrigger:fitCollisionMaskToAllowedVehicles()')
	
	local trigger_dynamic_objects = 16777216 -- bales, pallets, logs
	
	local collisionMask_old = getCollisionMask(self.triggerId)
	local collisionMask_new = collisionMask_old
	
	if (self.allowBales or self.allowPallets or self.allowObjects) and bitAND(collisionMask_new,trigger_dynamic_objects)==0 then
		self:printInfo('Warning: allowBales, allowPallets or allowObjects is set to true but collisionMask was not fitting (fixed)')
		collisionMask_new = collisionMask_new + trigger_dynamic_objects
	end
	
	-- result
	
	if collisionMask_new ~= collisionMask_old then
		self:printInfo('Warning: set collisionMask according to allowed vehicles to ',collisionMask_new,' (you may want to fix that)')
		setCollisionMask(self.triggerId,collisionMask_new)
	end
end

function UPK_ObjectTrigger:triggerCallback(triggerId, otherActorId, onEnter, onLeave, onStay, otherShapeId)
	-- overwrite default function
	self:printFn('UPK_ObjectTrigger:triggerCallback(',triggerId,', ',otherActorId,', ',onEnter,', ',onLeave,', ',onStay,', ',otherShapeId,')')
	if self.isEnabled then
		local object = g_currentMission.nodeObjects[otherShapeId] or
						g_currentMission.nodeObjects[otherActorId]
		self:printInfo('object is ',object)
		
		if object~=nil then
			if self.vehicleNodesInTrigger[object]==nil then
				self.vehicleNodesInTrigger[object]={}
			end
			
			self:printInfo('allowPallets ',self.allowPallets)
			self:printInfo('is pallet ',getIsPallet(object))
			self:printInfo('allowBales ',self.allowBales)
			self:printInfo('is bale ',getIsBale(object))
			self:printInfo('allowObjects ',self.allowObjects)
			self:printInfo('is object ',(object.isa~=nil and object:isa(self.objectClass)))
			
			if (self.allowPallets and getIsPallet(object)) or
				(self.allowBales and getIsBale(object)) or
				(self.allowObjects and object.isa~=nil and object:isa(self.objectClass)) then
				self:printInfo('object allowed')
				if onEnter then
					self:printInfo('onEnter')
					if not isInTable(self.vehicleNodesInTrigger[object],otherShapeId) then
						if self.deleteObjectOnEnter then
							object:delete()
						else
							table.insert(self.vehicleNodesInTrigger[object],otherShapeId)
							if length(self.vehicleNodesInTrigger[object])==1 then
								self:vehicleEntered(object) -- only for one node of the same object
							end
						end
					end
				elseif onLeave then
					self:printInfo('onLeave')
					removeValueFromTable(self.vehicleNodesInTrigger[object],otherShapeId)
					if length(self.vehicleNodesInTrigger[object])==0 then
						self.vehicleNodesInTrigger[object]=nil
						self:vehicleLeft(object)
						if self.deleteObjectOnLeave then
							object:delete()
						end
					end
				end
			end
		end
		--[[
		if self.allowWood then
			local splitType = SplitUtil.splitTypes[getSplitType(otherActorId)]
			self:printAll('splitType is ',splitType)
			if splitType~=nil and splitType.woodChipsPerLiter>0 then
				if onEnter then
					self:vehicleEntered(otherActorId,true)
				elseif onLeave then
					self:vehicleLeft(otherActorId,false)
				end
			end
		end
		]]--
	end
end

function UPK_ObjectTrigger:delete()
	self:printFn('UPK_ObjectTrigger:delete()')
	UPK_ObjectTrigger:superClass().delete(self)
end